#!/usr/local/bin/perl
# /etc/wwclient: Mount csd if present. It must be run on each csd client.
# INSTALL: Just type
#
#	wwclient -i
#
# then add the following lines to /etc/rc.local:
#
#	if [ -f /etc/wwclient -a -f /etc/wwfs.conf ]; then
#		/etc/wwclient
#	fi
#
# If you found any problem with csd, consult your local guru,
# presumably reachable at wwfs@ $cshost.
#
# Author: Youki Kadobayashi <youki@wide.ad.jp>
# This file is part of WWFS.
#
require 'getopts.pl';
do Getopts('i');

unless ($opt_i) {
	open(OUT, "/dev/console");
	select(OUT);
}

$conf = "/etc";
$skel = "/usr/local/ww/etc";

if (-e "$conf/wwfs.conf") {
	&readconf("$conf/wwfs.conf") || die "wwclient: $!\n";
} elsif ($opt_i) {
	&install;
} else {
	die "wwclient: $conf/wwfs.conf is missing\n";
}

$CS_PROGRAM=300128;	# hardcoded; do not change

# some sanity checks here
if ( ! -f $mountcmd || ! -x $mountcmd ) {
	print "wwfs.conf: \"$mountcmd\" is not executable\n";
	exit 1;
}
if ( ! -d $mountdir ) {
	if ($opt_i) {
		&install;
	} else {
		print "wwfs.conf: \"$mountdir\" is not a directory\n";
		exit 1;
	}
}

system("/usr/etc/rpcinfo -u $cshost $CS_PROGRAM");
if ($? == 0) {
	system("$mountcmd $cshost $mountdir");
} else {
	print "wwclient: csd not running on $cshost\n";
}
exit 0;

sub readconf {
	local ($file) = @_;
	open(CONF, $file) || return 0;
	while (<CONF>) {
		s/(\S+)=(\S+)$/\$\1="\2";/;
		eval $_;
	}
	close(CONF);
	1;
}

sub install {
	&readconf("$skel/wwfs.conf") || die "wwclient: $!\n";

	if (! -d $mountdir) {
		print "wwclient: creating mount point \"$mountdir\"\n";
		&mkdirhier($mountdir);
	}
	if (! -e "$conf/wwfs.conf") {
		print "wwclient: copying $skel/wwfs.conf to $conf\n";
		system("cp $skel/wwfs.conf $conf");
	}
	if (! -e "$conf/wwclient") {
		print "wwclient: copying $skel/wwclient to $conf\n";
		system("cp $skel/wwclient $conf");
	}
}

sub mkdirhier {
	local($dir) = @_;
	local($mk);
	# four-line implementation
	foreach $i (split (/\//, $dir)) {
		$mk .= "/" . $i;
		mkdir($mk, 0755);
	}
}
