/* 
 * WorldWide File System
 * Copyright (c) 1992,1993 Youki Kadobayashi
 * Copyright (c) 1992,1993 Osaka University
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that the following conditions
 * are met:
 * 1. Both the copyright notice and this permission notice appear in
 *    all copies of the software, derivative works or modified versions,
 *    and any portions thereof, and that both notices appear in
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Osaka University
 *      and its contributors.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE DEVELOPER ``AS IS'' AND OSAKA
 * UNIVERSITY DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
 * WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Osaka University requests users of this software to return to
 *
 *  Youki Kadobayashi
 *  Department of Information and Computer Sciences
 *  Osaka University, Toyonaka 560, Osaka, Japan
 *
 * any improvements or extensions that they make and grant Osaka
 * University the rights to redistribute these changes.
 */
static char *AtFSid = "$Header: conn.c[109.0] Wed Nov 24 03:47:04 1993 youki-k@is.aist-nara.ac.jp saved $";

#include "wfs.h"
#include "util.h"
#include "global.h"

extern qelem conn_head;
qelem conn_head = { &conn_head, &conn_head };
static void *conn_callout;

static void conn_gc(), conn_close();

void
conn_start()
{
	conn_callout = timeout_set(gettime()+60, conn_gc, 0);
}

static void
conn_gc()
{
	wf_conn *p, *p2;

	ITER2(p, p2, wf_conn, &conn_head) {
		if (p->thrdp) continue;
		p->idle += 60;
		if (p->idle > WF_CONN_TTL) {
			conn_close(p);
		}
	}
	conn_callout = timeout_set(gettime()+60, conn_gc, 0);
}

void
conn_stop()
{
	wf_conn *cp;
	ITER(cp, wf_conn, &conn_head) {
		conn_close(cp);
	}
}

int
conn_count()
{
	return q_len(&conn_head);
}

wf_conn *
conn_alloc()
{
	wf_conn *cp;

	cp = ALLOC(wf_conn);
	cp->proto = &proto_ftp;		/* XXX */
	q_insert(cp, &conn_head);
	return cp;
}

void
conn_free(cp)
wf_conn *cp;
{
	if (cp->id == 0) {
		syslog(LOG_WARNING, "connp->id == 0");
	}
	q_remove(cp);
	FREE(cp);
}

static void
conn_close(cp)
wf_conn *cp;
{
	srv_releaseconn(cp->srv, cp);
	proto_close(cp);
	conn_free(cp);
}

void
conn_shutdown(cp, howto)
wf_conn *cp;
int howto;
{
	srv_releaseconn(cp->srv, cp);
	proto_shutdown(cp, howto);
	conn_free(cp);
}

void
conn_dropped(so)
int so;
{
	wf_conn *p;

	ITER(p, wf_conn, &conn_head) {
		if (p->id == so) {
			conn_close(p, 2);
			return;
		}
	}
	dlog("conn_dropped: no such connection");
}

void
conn_inspect(reqp)
wf_req *reqp;
{
	wf_conn *p;

	ITER(p, wf_conn, &conn_head) {
		req_send(reqp, "conn %lx: id= %d\n", p, p->id);
		req_send(reqp, "\tidle= %d, srv=%s, thrdp=%lx\n",
			 p->idle, PROT(p->srv->name), p->thrdp);
	}
}
