/* 
 * WorldWide File System
 * Copyright (c) 1992,1993 Youki Kadobayashi
 * Copyright (c) 1992,1993 Osaka University
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that the following conditions
 * are met:
 * 1. Both the copyright notice and this permission notice appear in
 *    all copies of the software, derivative works or modified versions,
 *    and any portions thereof, and that both notices appear in
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Osaka University
 *      and its contributors.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE DEVELOPER ``AS IS'' AND OSAKA
 * UNIVERSITY DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
 * WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Osaka University requests users of this software to return to
 *
 *  Youki Kadobayashi
 *  Department of Information and Computer Sciences
 *  Osaka University, Toyonaka 560, Osaka, Japan
 *
 * any improvements or extensions that they make and grant Osaka
 * University the rights to redistribute these changes.
 */
/* NFS reply */
static char *AtFSid = "$Header: nfs_reply.c[109.0] Wed Nov 24 03:47:13 1993 youki-k@is.aist-nara.ac.jp saved $";

#include <sys/file.h>		/* for O_RDONLY */
#include "wfs.h"
#include "util.h"
#include "global.h"

/* translation of WFS semantics to NFS errno */
static nfsstat nfserr[] = {
	NFS_OK,		/* WF_REP_OK: no error */
	NFSERR_PERM,	/* WF_ERR_PERM: Not owner */
	NFSERR_NOENT,	/* WF_ERR_NOENT: No such entity */
	NFSERR_IO,	/* WF_ERR_IO: I/O error */
	NFSERR_NXIO,	/* WF_ERR_NXIO: No such device or address */
	NFSERR_ACCES,	/* WF_ERR_ACCES: Permission denied */
	NFSERR_EXIST,	/* WF_ERR_EXIST: Given resource exists */
	NFSERR_NODEV,	/* WF_ERR_NODEV: No such device */
	NFSERR_NOTDIR,	/* WF_ERR_NOTDIR: Not a directory*/
	NFSERR_ISDIR,	/* WF_ERR_ISDIR: Is a directory */
	NFSERR_FBIG,	/* WF_ERR_FBIG: Resource too large */
	NFSERR_NOSPC,	/* WF_ERR_NOSPC: No space left on device */
	NFSERR_ROFS,	/* WF_ERR_ROFS: Read-only file system */
	NFSERR_NAMETOOLONG, /* WF_ERR_NAMETOOLONG: Resource name too long */
	NFSERR_NOTEMPTY, /* WF_ERR_NOTEMPTY: Directory not empty */
	NFSERR_DQUOT,	/* WF_ERR_DQUOT: Disc quota exceeded */
	NFSERR_STALE,	/* WF_ERR_STALE: Stale file handle */
	NFSERR_WFLUSH,	/* WF_ERR_WFLUSH: Write cache flushed */
	/* brain damaged NFS does not know what Internet is */
	NFSERR_NOENT,	/* WF_REP_WAIT: Operation would block */
	NFSERR_NOENT,	/* WF_ERR_ALREADY: Operation already in progress */
	NFSERR_NOENT,	/* WF_ERR_PROTO: Protocol not available */
	NFSERR_NOENT,	/* WF_ERR_NDOWN: Network is down */
	NFSERR_NOENT,	/* WF_ERR_NUNREACH: Network is unreachable */
	NFSERR_NOENT,	/* WF_ERR_XABORT: Transfer aborted */
	NFSERR_NOENT,	/* WF_ERR_TIMEOUT: Connection timed out */
	NFSERR_NOENT,	/* WF_ERR_REFUSED: Connection refused */
	NFSERR_NOENT,	/* WF_ERR_HDOWN: Host is down */
	NFSERR_NOENT,	/* WF_ERR_HUNREACH: No route to host */
	NFSERR_NOENT,	/* WF_ERR_NQUOT: Network quota exceeded */
};

void
nfs_getattr_reply(c, status)
wf_thrd *c;
int status;
{
	static attrstat res;
	if (status == WF_REP_OK) {
		bcopy(&c->filep->fattr, &res.attrstat_u.attributes,
		      sizeof(fattr));
	} else {
		bzero(&res, sizeof (res));
	}
	res.status = nfserr[status];
	svc_sendreply(c->rqstp->rq_xprt, xdr_attrstat, &res);
	if (! (c->flag & WF_THREAD_BUSY)) thrd_free(c);
}

void
nfs_lookup_reply(c, status)
wf_thrd	*c;
int status;
{
	static diropres res;
	if (status == WF_REP_OK) {
		bcopy(c->resfhp, &res.diropres_u.diropres.file,
		      sizeof (nfs_fh));
		bcopy(&c->filep->fattr, &res.diropres_u.diropres.attributes,
		      sizeof(fattr));
	} else {
		bzero(&res, sizeof (res));
	}
	res.status = nfserr[status];
	svc_sendreply(c->rqstp->rq_xprt, xdr_diropres, &res);
	if (! (c->flag & WF_THREAD_BUSY)) thrd_free(c);
}

void
nfs_readlink_reply(c, status)
wf_thrd *c;
int status;
{
	static readlinkres res;
	char path[NFS_MAXPATHLEN];
	if (status == WF_REP_OK) {
		res.readlinkres_u.data = path;
		strcpy(res.readlinkres_u.data, c->filep->link);
	} else {
		bzero(&res, sizeof (res));
	}
	res.status = nfserr[status];
	svc_sendreply(c->rqstp->rq_xprt, xdr_readlinkres, &res);
	if (! (c->flag & WF_THREAD_BUSY)) thrd_free(c);
}

void
nfs_read_reply(c, status, attrp, data_len, data_val)
wf_thrd *c;
nfsstat status;
fattr *attrp;
int data_len;
char *data_val;
{
	static readres res;
	if (attrp && data_val) {
		bcopy(attrp, &res.readres_u.reply.attributes, sizeof(fattr));
		res.readres_u.reply.data.data_len = data_len;
		res.readres_u.reply.data.data_val = data_val;
	} else {
		bzero(&res, sizeof (res));
	}
	res.status = nfserr[status];
	svc_sendreply(c->rqstp->rq_xprt, xdr_readres, &res);
	if (! (c->flag & WF_THREAD_BUSY)) thrd_free(c);
}

static void
nfs_readdir_reply(c, status, resp)
wf_thrd *c;
nfsstat status;
readdirres *resp;
{
	static readdirres res;

	if (resp) {
		resp->status = status;
		svc_sendreply(c->rqstp->rq_xprt, xdr_readdirres, resp);
	} else {
		bzero(&res, sizeof(res));
		res.status = status;
		svc_sendreply(c->rqstp->rq_xprt, xdr_readdirres, &res);
	}
	if (! (c->flag & WF_THREAD_BUSY)) thrd_free(c);
}


void
nfs_read_done(c, status)
/* rqstp, volp, offset, count, totalcount, dirp, filep */
wf_thrd *c;
int status;
{
	int fd, buflen;
	char *buf;
	char path[MAXPATHLEN];

	if (status != WF_REP_OK) {
		nfs_read_reply(c, status, 0, 0, 0);
		return;
	}
	file_getpath(path, c->child_volp, c->dirp, c->filep);
	fd = open(path, O_RDONLY);
	if (fd < 0) {
		dlog("no such file: %s", path);
		nfs_read_reply(c, WF_ERR_NOENT, 0, 0, 0);
		return;
	}
	lseek(fd, (off_t )c->offset, 0);
	buflen = c->count;
	buf = so_getbuf(fd);
	buflen = read(fd, buf, buflen);
	close(fd);
	if (buflen < 0) {
		dlog("cmfs_read: read error");
		nfs_read_reply(c, WF_ERR_IO, 0, 0, 0);
		return;
	}
	c->child_volp->stats.bytes_to_client += buflen;
	nfs_read_reply(c, WF_REP_OK, &c->filep->fattr, buflen, buf);
}

#ifdef DEBUG_CHAIN
static void
check_chain(c, filep, ident)
wf_thrd *c;
wf_file *filep;
char *ident;
{
	wf_file *d;
	int found = 0;

	if (filep == c->dirp->file)
		return;
	ITER(d, wf_file, &c->dirp->file->q) {
		if (d == filep) {
			found = 1;
			break;
		}
	}
	if (! found) {
		dlog("%s: filep = %08x not in chain", ident, filep);
		csd_abort(1);
	}
}
#endif

/*
 * You cannot eat these cookies.
 */
static unsigned int	gen = 1;	/* generation number */
static unsigned int	ch;		/* cookie handle */
#define	CH_SIZE	256

/* Cookie internals */
union cookie {
	struct {
		unsigned int	gen: 24;
		unsigned int	ch: 8;
	} broken;
	unsigned long whole;
};

/* Ring buffer of cookie checker/mapper. Indexed by cookie handle */
static struct cookie_checker {
	wf_file	*filep;
	wf_file	*head;
} cooche[CH_SIZE];

unsigned long
cookie_bake(filep, head)
wf_file *filep, *head;
{
	union cookie u;

	assert(sizeof(union cookie) == sizeof(nfscookie));
	if (++ch == CH_SIZE) {
		ch = 0;
		++gen;
		if (gen == 0)
			gen = 1;	/* so that valid cookie != 0 holds */
	}
	cooche[ch].filep = filep;
	cooche[ch].head = head;
	u.broken.gen = gen;
	u.broken.ch = ch;
	return (u.whole);
}

wf_file *
cookie_taste(cookie, head)
unsigned long cookie;
wf_file *head;
{
	unsigned int i;
	union cookie u;

	u.whole = cookie;
	i = u.broken.ch;
	if (u.broken.gen == gen || (u.broken.gen == gen-1 && i > ch)) {
		if (cooche[i].head == head)
			return (cooche[i].filep);
	}
	return (wf_file *)0;
}

void
nfs_readdir_done(c, status)
wf_thrd *c;	/* rqstp, volp, cookie, count, dirp */
int status;
{
	static readdirres res;
	wf_file *filep, *nextp;
	entry **e, *p, *p2;
	unsigned long cool;
	int size = 0;

	if (status != WF_REP_OK) {
		nfs_readdir_reply(c, nfserr[status], 0);
		return;
	}
	e = &res.readdirres_u.reply.entries;
	bcopy(c->cookie, &cool, sizeof (nfscookie));
	if (! cool) {
		filep = FIRST(wf_file, &c->dirp->file->q);
	} else {
		filep = cookie_taste(cool, c->dirp->file);
		if (! filep) {
			struct in_addr in;
			in.s_addr = c->client;
			dlog("invalid cookie from client %s",
			     inet_ntoa(in));
			nfs_readdir_reply(c, NFSERR_NOENT, 0);
			return;
		}
#ifdef DEBUG_CHAIN
		check_chain(c, filep, "get cookie");
#endif
	}
	for (; nextp = NEXT(wf_file, filep),
	     filep != HEAD(wf_file, &c->dirp->file->q);
	     filep = nextp) {
		size += sizeof(entry) + strlen(filep->name) + 16;
		if (size >= c->count)
			break;
		*e = ALLOC(entry);
		(*e)->fileid = filep->id;
		(*e)->name = filep->name;
		cool = cookie_bake(nextp, c->dirp->file);
		bcopy(&cool, (*e)->cookie, sizeof(nfscookie));
#ifdef DEBUG_CHAIN
		check_chain(c, nextp, "set cookie");
#endif
		e = &((*e)->nextentry);
	}
	*e = (entry *)0;
	res.readdirres_u.reply.eof =
		(filep == HEAD(wf_file, &c->dirp->file->q));
	c->child_volp->stats.bytes_to_client += MIN(size, c->count);
	nfs_readdir_reply(c, NFS_OK, &res);

	p = res.readdirres_u.reply.entries;
	while (p) {
		p2 = p->nextentry;
		FREE(p);
		p = p2;
	}
}
