/* 
 * WorldWide File System
 * Copyright (c) 1992,1993 Youki Kadobayashi
 * Copyright (c) 1992,1993 Osaka University
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that the following conditions
 * are met:
 * 1. Both the copyright notice and this permission notice appear in
 *    all copies of the software, derivative works or modified versions,
 *    and any portions thereof, and that both notices appear in
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Osaka University
 *      and its contributors.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE DEVELOPER ``AS IS'' AND OSAKA
 * UNIVERSITY DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
 * WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Osaka University requests users of this software to return to
 *
 *  Youki Kadobayashi
 *  Department of Information and Computer Sciences
 *  Osaka University, Toyonaka 560, Osaka, Japan
 *
 * any improvements or extensions that they make and grant Osaka
 * University the rights to redistribute these changes.
 */
/* nfsproc_whatever_2 */
static char *AtFSid = "$Header: nfs_subr.c[109.0] Wed Nov 24 03:47:14 1993 youki-k@is.aist-nara.ac.jp saved $";

#include "wfs.h"
#include "util.h"
#include "global.h"

#define NFS
#define NFSCLIENT
#ifdef NFS_3
typedef nfs_fh fhandle_t;
#endif /* NFS_3 */
#include <sys/mount.h>		/* for struct statfs */
#ifdef STATFS_HDR
#include STATFS_HDR		/* for struct statfs */
#endif /* STATFS_HDR */

static void
client_info(c)
wf_thrd *c;
{
	struct authunix_parms *unix_cred;
	sockaddr_in *sin;

	if (c->rqstp->rq_cred.oa_flavor == AUTH_UNIX) {
		unix_cred = (struct authunix_parms *) c->rqstp->rq_clntcred;
		c->uid = unix_cred->aup_uid;
	}
	sin = svc_getcaller(c->rqstp->rq_xprt);
	c->client = sin->sin_addr.s_addr;
}

void *
nfsproc_null_2(argp, rqstp)
void *argp;
svc_req *rqstp;
{
	static char res;
	return (void *) &res;
}

attrstat *
nfsproc_getattr_2(argp, rqstp)
nfs_fh *argp;
svc_req *rqstp;
{
	wf_thrd *c;
	static attrstat res;
	wf_fh *fhp = (wf_fh *)argp;
	wf_vol *parent_volp, *child_volp;

	/* auth */
	if (fhp->world_id != cs_world) {
		bzero(&res, sizeof(res));
		res.status = NFSERR_ACCES;
		return &res;
	}

	parent_volp = vol_findid(fhp->parent_vol);
	if (fhp->parent_vol == fhp->child_vol)
		child_volp = parent_volp;
	else
		child_volp = vol_findid(fhp->child_vol);
	if (parent_volp == NULL || child_volp == NULL) {
		bzero(&res, sizeof(res));
		res.status = NFSERR_STALE;
		return &res;
	}
	c = thrd_alloc();
	c->rqstp = CLONE(rqstp);
	c->parent_volp = parent_volp;
	c->child_volp = child_volp;
	c->fhp = fhp;
	c->reply = nfs_getattr_reply;
	proto_getattr(c);
	return (attrstat *)0;
}

attrstat *
nfsproc_setattr_2(argp, rqstp)
sattrargs *argp;
svc_req *rqstp;
{
	static attrstat res;
	bzero(&res, sizeof(res));
	res.status = NFSERR_ROFS;
	return &res;
}

void *
nfsproc_root_2(argp, rqstp)
void *argp;
svc_req *rqstp;
{
	static char res;
	return (void *) &res;
}

diropres *
nfsproc_lookup_2(argp, rqstp)
diropargs *argp;
svc_req *rqstp;
{
	wf_thrd *c;
	static diropres res;
	wf_fh *fhp = (wf_fh *) &argp->dir;
	wf_vol *parent_volp, *child_volp;

	/* auth */
	if (fhp->world_id != cs_world) {
		bzero(&res, sizeof(res));
		res.status = NFSERR_ACCES;
		return &res;
	}

	parent_volp = vol_findid(fhp->parent_vol);
	if (fhp->parent_vol == fhp->child_vol)
		child_volp = parent_volp;
	else
		child_volp = vol_findid(fhp->child_vol);
	if (parent_volp == NULL || child_volp == NULL) {
		bzero(&res, sizeof(res));
		res.status = NFSERR_STALE;
		return &res;
	}
	c = thrd_alloc();
	c->rqstp = CLONE(rqstp);
	c->parent_volp = parent_volp;
	c->child_volp = child_volp;
	c->fhp = fhp;
	c->fname = argp->name;
	c->reply = nfs_lookup_reply;
	client_info(c);
	/*
	 * gross hack
	 */
	if (! (c->fname[0] == '.' && strncmp(c->fname, ".wwfs_", 6) == 0)) {
		proto_lookup(c);
		return (diropres *)0;
	} else {
		wf_fh *resfhp;
		int ret;
		/*
		 * Strict specification of NFS was successful in that it
		 * prohibited vendors to add diverse features.
		 * Meanwhile, the strictness necessitates brute force
		 * solution, which is much more uglier than ioctl()...
		 * I have long-term solution in mind however.
		 */
		c->resfhp = resfhp = CLONE(fhp);
#if 0
		if (strcmp(c->fname + 6, "host") == 0) {
			resfhp->child_vol = WF_HOST_VOL_ID;
		} else if (strcmp(c->fname + 6, "cwd") == 0) {
			resfhp->child_vol = WF_CWD_VOL_ID;
		}
#else
		if (strcmp(c->fname + 6, "fh") == 0) {
			resfhp->child_vol = WF_FH_VOL_ID;
		}
#endif
		else {
			FREE(resfhp);
			bzero(&res, sizeof(res));
			res.status = NFSERR_NOENT;
			return &res;
		}
		resfhp->parent_vol = fhp->child_vol;
		ret = cwd_fake(c, resfhp);
		nfs_lookup_reply(c, ret);
		FREE(resfhp);
		return (diropres *)0;
	}
}

readlinkres *
nfsproc_readlink_2(argp, rqstp)
nfs_fh *argp;
svc_req *rqstp;
{
	wf_thrd *c;
	static readlinkres res;
	wf_fh *fhp = (wf_fh *)argp;
	wf_vol *parent_volp, *child_volp;

	/* auth */
	if (fhp->world_id != cs_world) {
		bzero(&res, sizeof(res));
		res.status = NFSERR_ACCES;
		return &res;
	}

	parent_volp = vol_findid(fhp->parent_vol);
	if (fhp->parent_vol == fhp->child_vol)
		child_volp = parent_volp;
	else
		child_volp = vol_findid(fhp->child_vol);
	if (parent_volp == NULL || child_volp == NULL) {
		bzero(&res, sizeof(res));
		res.status = NFSERR_STALE;
		return &res;
	}
	c = thrd_alloc();
	c->rqstp = CLONE(rqstp);
	c->parent_volp = parent_volp;
	c->child_volp = child_volp;
	c->fhp = fhp;
	c->reply = nfs_readlink_reply;
	proto_readlink(c);
	return (readlinkres *)0;
}

readres *
nfsproc_read_2(argp, rqstp)
readargs *argp;
svc_req *rqstp;
{
	wf_thrd *c;
	static readres res;
	wf_fh *fhp = (wf_fh *) &argp->file;
	wf_vol *parent_volp, *child_volp;

	/* auth */
	if (fhp->world_id != cs_world) {
		bzero(&res, sizeof(res));
		res.status = NFSERR_ACCES;
		return &res;
	}

	parent_volp = vol_findid(fhp->parent_vol);
	if (fhp->parent_vol == fhp->child_vol)
		child_volp = parent_volp;
	else
		child_volp = vol_findid(fhp->child_vol);
	if (parent_volp == NULL || child_volp == NULL) {
		bzero(&res, sizeof(res));
		res.status = NFSERR_STALE;
		return &res;
	}
	c = thrd_alloc();
	c->rqstp = CLONE(rqstp);
	c->parent_volp = parent_volp;
	c->child_volp = child_volp;
	c->fhp = fhp;
	c->offset = argp->offset;
	c->count = argp->count;
	c->totalcount = argp->totalcount;
	c->reply = nfs_read_done;
	client_info(c);
	proto_read(c);
	return (readres *)0;
}

void *
nfsproc_writecache_2(argp, rqstp)
void *argp;
svc_req *rqstp;
{
	static char res;
	return (void *)&res;
}

attrstat *
nfsproc_write_2(argp, rqstp)
writeargs *argp;
svc_req *rqstp;
{
	static attrstat res;
	bzero(&res, sizeof(res));
	res.status = NFSERR_ROFS;
	return &res;
}

diropres *
nfsproc_create_2(argp, rqstp)
createargs *argp;
svc_req *rqstp;
{
	static diropres res;
	bzero(&res, sizeof(res));
	res.status = NFSERR_ROFS;
	return &res;
}

nfsstat *
nfsproc_remove_2(argp, rqstp)
diropargs *argp;
svc_req *rqstp;
{
	static nfsstat res;
	bzero(&res, sizeof(res));
	res = NFSERR_ROFS;
	return &res;
}

nfsstat *
nfsproc_rename_2(argp, rqstp)
renameargs *argp;
svc_req *rqstp;
{
	static nfsstat res;
	bzero(&res, sizeof(res));
	res = NFSERR_ROFS;
	return &res;
}

nfsstat *
nfsproc_link_2(argp, rqstp)
linkargs *argp;
svc_req *rqstp;
{
	static nfsstat res;
	bzero(&res, sizeof(res));
	res = NFSERR_ROFS;
	return &res;
}

nfsstat *
nfsproc_symlink_2(argp, rqstp)
symlinkargs *argp;
svc_req *rqstp;
{
	static nfsstat res;
	bzero(&res, sizeof(res));
	res = NFSERR_ROFS;
	return &res;
}

diropres *
nfsproc_mkdir_2(argp, rqstp)
createargs *argp;
svc_req *rqstp;
{
	static diropres res;
	bzero(&res, sizeof(res));
	res.status = NFSERR_ROFS;
	return &res;
}

nfsstat *
nfsproc_rmdir_2(argp, rqstp)
diropargs *argp;
svc_req *rqstp;
{
	static nfsstat res;
	bzero(&res, sizeof(res));
	res = NFSERR_ROFS;
	return &res;
}

readdirres *
nfsproc_readdir_2(argp, rqstp)
readdirargs *argp;
svc_req *rqstp;
{
	wf_thrd *c;
	static readdirres res;
	wf_fh *fhp = (wf_fh *) &argp->dir;
	wf_vol *parent_volp, *child_volp;

	bzero(&res, sizeof(res));
	/* auth */
	if (fhp->world_id != cs_world) {
		res.status = NFSERR_ACCES;
		return &res;
	}

	parent_volp = vol_findid(fhp->parent_vol);
	if (fhp->parent_vol == fhp->child_vol)
		child_volp = parent_volp;
	else
		child_volp = vol_findid(fhp->child_vol);
	if (parent_volp == NULL || child_volp == NULL) {
		bzero(&res, sizeof(res));
		res.status = NFSERR_STALE;
		return &res;
	}
	c = thrd_alloc();
	c->rqstp = CLONE(rqstp);
	c->parent_volp = parent_volp;
	c->child_volp = child_volp;
	c->fhp = fhp;
	bcopy(argp->cookie, c->cookie, sizeof(nfscookie));
	c->count = argp->count;
	c->reply = nfs_readdir_done;
	client_info(c);
	/*
	 * access control (only when access control file is present)
	 */
	if (parent_volp->accp
	    && acc_check(parent_volp->accp, c->client, WF_ACC_NFS) < 0) {
		(*c->reply)(c, WF_ERR_ACCES);
		return (readdirres *)0;
	}
	proto_readdir(c);
	return (readdirres *)0;
}

statfsres *
nfsproc_statfs_2(argp, rqstp)
nfs_fh *argp;
svc_req *rqstp;
{
	static statfsres res;
	wf_fh *fhp = (wf_fh *) argp;
#ifdef ULTRIX_HACK
	struct fs_data buf;
#else /* ULTRIX_HACK */
	struct statfs buf;
#endif /* ULTRIX_HACK */

	bzero(&res, sizeof(res));
	/* auth */
	if (fhp->world_id != cs_world) {
		res.status = NFSERR_ACCES;
		return &res;
	}

	/* statfs cache partition */
	statfs(cs_topdir, &buf);
	res.status = NFS_OK;
	res.statfsres_u.reply.tsize = NFS_MAXDATA;
#ifdef ULTRIX_HACK
	res.statfsres_u.reply.bsize = buf.fd_bsize;
	res.statfsres_u.reply.blocks = buf.fd_btot / (buf.fd_bsize / 1024);
	res.statfsres_u.reply.bfree = buf.fd_bfree / (buf.fd_bsize / 1024);
	res.statfsres_u.reply.bavail = buf.fd_bfreen / (buf.fd_bsize / 1024);
#else /* ULTRIX_HACK */
	res.statfsres_u.reply.bsize = buf.f_bsize;
	res.statfsres_u.reply.blocks = buf.f_blocks;
	res.statfsres_u.reply.bfree = buf.f_bfree;
#ifdef sgi	/* or HAS_BAVAIL? */
	/* return estimated 95% figure */
	res.statfsres_u.reply.bavail = buf.f_bfree - (buf.f_bfree / 20);
#else /* sgi */
	res.statfsres_u.reply.bavail = buf.f_bavail;
#endif /* sgi */
#endif /* ULTRIX_HACK */
	return (statfsres *)&res;
}

