static char *AtFSid = "$Header: sched.c[109.0] Wed Nov 24 03:47:16 1993 youki-k@is.aist-nara.ac.jp saved $";
/*
 * Copyright (c) 1990 Jan-Simon Pendry
 * Copyright (c) 1990 Imperial College of Science, Technology & Medicine
 * Copyright (c) 1990 The Regents of the University of California.
 * All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Jan-Simon Pendry at Imperial College, London.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	%W% (Berkeley) %G%
 *
 * $Id: sched.c,v 5.2.2.1 1992/02/09 15:09:02 jsp beta $
 *
 */

/*
 * Process scheduler
 */
/* #define DEBUG_SLEEP */

#include <sys/signal.h>
#include <sys/wait.h>
#include <setjmp.h>
#include "wfs.h"
#include "util.h"
extern jmp_buf select_intr;
extern int select_intr_valid;

typedef struct pjob pjob;
struct pjob {
	qelem hdr;			/* Linked list */
	int pid;			/* Process ID of job */
	func cb_fun;			/* Callback function */
	void *cb_closure;		/* Closure for callback */
	void *wchan;			/* Wait channel */
};

extern qelem proc_list_head;
qelem proc_list_head = { &proc_list_head, &proc_list_head };
extern qelem proc_wait_list;
qelem proc_wait_list = { &proc_wait_list, &proc_wait_list };

int task_notify_todo;

static pjob *sched_job(cf, ca)
func cf;
void *ca;
{
	pjob *p = ALLOC(pjob);

	p->cb_fun = cf;
	p->cb_closure = ca;

	/* Now append on wait queue */
	q_insert(p, LAST(pjob, &proc_wait_list));

	return p;
}

static pjob *sched_job_atop(cf, ca)
func cf;
void *ca;
{
	pjob *p = ALLOC(pjob);

	p->cb_fun = cf;
	p->cb_closure = ca;

	/* Now place on top of wait queue */
	q_insert(p, &proc_wait_list);

	return p;
}

void run_task(tf, ta, cf, ca)
int (*tf)();
void *ta;
func cf;
void *ca;
{
	pjob *p = sched_job(cf, ca);
	int mask;

	p->wchan = (void *) p;

	mask = sigblock(sigmask(SIGCHLD));

	if (p->pid = background()) {
		sigsetmask(mask);
		return;
	}

	exit((*tf)(ta));
	/* firewall... */
	abort();
}

/*
 * Schedule a task to be run when woken up
 */
void sched_task(cf, ca, wchan)
func cf;
void *ca;
void *wchan;
{
	/*
	 * Allocate a new task
	 */
	pjob *p = sched_job(cf, ca);
	p->wchan = wchan;
	p->pid = 0;
#ifdef DEBUG_SLEEP
	dlog("SLEEP on %x", wchan);
#endif
}

/*
 * On some specific situation, the order of wakeupjob() calls
 * lead to a problem, since every resource manipulation functions
 * should be bracketed between resource-allocation functions and
 * resource-freeing functions; incorrect ordering of wakeupjob()
 * will violate this rule.
 *
 * sched_task_atop() was devised for this purpose; i.e. to avoid
 * freeing resources before they are accessed.
 */
void sched_task_atop(cf, ca, wchan)
func cf;
void *ca;
void *wchan;
{
	/*
	 * Allocate a new task
	 */
	pjob *p = sched_job_atop(cf, ca);
	p->wchan = wchan;
	p->pid = 0;
#ifdef DEBUG_SLEEP
	dlog("SLEEP on %x", wchan);
#endif
}

static void wakeupjob(p)
pjob *p;
{
	q_remove(p);
	q_insert(p, &proc_list_head);
	task_notify_todo++;
#ifdef DEBUG_SLEEP
	dlog("task_notify_todo = %d", task_notify_todo);
#endif
}

void wakeup(wchan)
void *wchan;
{
	pjob *p, *p2;
#ifdef DEBUG_SLEEP
	int done = 0;
#endif

#ifdef DEBUG_SLEEP
	dlog("wakeup(%x)", wchan);
#endif
	/* Use ITER2() here because wakeupjob() juggles the list. */
	ITER2(p, p2, pjob, &proc_wait_list) {
		if (p->wchan == wchan) {
#ifdef DEBUG_SLEEP
			done = 1;
#endif
			wakeupjob(p);
		}
	}

#ifdef DEBUG_SLEEP
	if (!done)
		dlog("Nothing SLEEPing on %x", wchan);
#endif
}

/*ARGSUSED*/

void sigchld(sig)
int sig;
{
	union wait w;
	int pid;

#ifdef SYS5_SIGNALS
	if ((pid = wait(&w)) > 0) {
#else
	while ((pid = wait3((int *) &w, WNOHANG, (struct rusage *) 0)) > 0) {
#endif /* SYS5_SIGNALS */
		pjob *p, *p2;

		if (WIFSIGNALED(w))
			syslog(LOG_ERR, "Process %d exited with signal %d",
				pid, w.w_termsig);
#ifdef DEBUG
		else
			dlog("Process %d exited with status %d",
				pid, w.w_retcode);
#endif /* DEBUG */

		ITER2(p, p2, pjob, &proc_wait_list) {
			if (p->pid == pid) {
				wakeupjob(p);
				break;
			}
		}

#ifdef DEBUG
		if (p) ; else dlog("can't locate task block for pid %d", pid);
#endif /* DEBUG */
	}

#ifdef SYS5_SIGNALS
	signal(sig, sigchld);
#endif /* SYS5_SIGNALS */
	if (select_intr_valid)
		longjmp(select_intr, sig);
}

/*
 * Run any pending tasks.
 * This must be called with SIGCHLD disabled
 */
void do_task_notify()
{
	/*
	 * Keep taking the first item off the list and processing it.
	 *
	 * Done this way because the the callback can, quite reasonably,
	 * queue a new task, so no local reference into the list can be
	 * held here.
	 */
#ifdef DEBUG_SLEEP
	dlog("do_task_notify");
#endif
	while (MORE(&proc_list_head)) {
		pjob *p = FIRST(pjob, &proc_list_head);
		q_remove(p);
		/*
		 * This job has completed
		 */
		--task_notify_todo;
#ifdef DEBUG_SLEEP
		dlog("task_notify_todo = %d", task_notify_todo);
#endif

		/*
		 * Do callback if it exists
		 */
		if (p->cb_fun) {
			(*p->cb_fun)(p->cb_closure);
		}

		FREE((void *) p);
	}
}

#ifdef HAS_SVR3_SIGNALS
/*
 * 4.2 signal library based on svr3 (4.1+ bsd) interface
 * From Stephen C. Pope <scp@acl.lanl.gov).
 */

static int current_mask = 0;

int sigblock(mask)
int mask;
{
    int sig;
    int m;
    int oldmask;

    oldmask = current_mask;
    for ( sig = 1, m = 1; sig <= MAXSIG; sig++, m <<= 1 ) {
        if (mask & m)  {
	    sighold(sig);
            current_mask |= m;
        }
    }
    return oldmask;
}

int sigsetmask(mask)
int mask;
{
    int sig;
    int m;
    int oldmask;

    oldmask = current_mask;
    for ( sig = 1, m = 1; sig <= MAXSIG; sig++, m <<= 1 ) {
        if (mask & m)  {
            sighold(sig);
            current_mask |= m;
        }
        else  {
            sigrelse(sig);
            current_mask &= ~m;
        }
    }
    return oldmask;
}

#endif /* HAS_SVR3_SIGNALS */
