/* 
 * WorldWide File System
 * Copyright (c) 1992,1993 Youki Kadobayashi
 * Copyright (c) 1992,1993 Osaka University
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that the following conditions
 * are met:
 * 1. Both the copyright notice and this permission notice appear in
 *    all copies of the software, derivative works or modified versions,
 *    and any portions thereof, and that both notices appear in
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Osaka University
 *      and its contributors.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE DEVELOPER ``AS IS'' AND OSAKA
 * UNIVERSITY DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
 * WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Osaka University requests users of this software to return to
 *
 *  Youki Kadobayashi
 *  Department of Information and Computer Sciences
 *  Osaka University, Toyonaka 560, Osaka, Japan
 *
 * any improvements or extensions that they make and grant Osaka
 * University the rights to redistribute these changes.
 */
/* thread status manager */
static char *AtFSid = "$Header: thrd.c[109.0] Wed Nov 24 03:47:17 1993 youki-k@is.aist-nara.ac.jp saved $";

#include "wfs.h"
#include "util.h"
#include "global.h"

extern qelem thrd_head;
qelem thrd_head = { &thrd_head, &thrd_head };
static void *thrd_callout;

wf_thrd *
thrd_alloc()
{
	wf_thrd *thrdp;

	thrdp = ALLOC(wf_thrd);
	return thrdp;
}

void
thrd_watch(thrdp)
wf_thrd *thrdp;
{
	q_insert(thrdp, &thrd_head);
}

void
thrd_ignore(thrdp)
wf_thrd *thrdp;
{
	q_remove(thrdp);
}

void
thrd_free(thrdp)
wf_thrd *thrdp;
{
	if (NEXT(wf_thrd, thrdp)) q_remove(thrdp);
	if (thrdp->rqstp) FREE(thrdp->rqstp);
	FREE(thrdp);
}

void
thrd_srvfailure(srvp)
wf_srv *srvp;
{
	/* server failure */
	wf_thrd *p, *p2;

	ITER2(p, p2, wf_thrd, &thrd_head) {
		if (p->cp && p->cp->srv == srvp) {
			(*p->give_up)(p->closure);
		}
	}
}

static void
thrd_gc()
{
	wf_thrd *p, *p2;

	ITER2(p, p2, wf_thrd, &thrd_head) {
		if (p->flag & WF_THREAD_ERROR) {
			(*p->give_up)(p->closure);
			p->flag &= ~WF_THREAD_ERROR;	/* called only once */
		} else if ((p->flag & WF_THREAD_RISKY)
			   && p->idle > WF_THREAD_TTL) {
			if (p->cp)
				srv_flag(p->cp->srv, WF_SRV_TIMEOUT);
			(*p->give_up)(p->closure);
			p->flag &= ~WF_THREAD_RISKY;	/* called only once */
		}
		if (p->flag & WF_THREAD_RISKY)
			p->idle += 60;
	}
	thrd_callout = timeout_set(gettime()+60, thrd_gc, 0);
}

void
thrd_start()
{
	thrd_callout = timeout_set(gettime()+60, thrd_gc, 0);
}

void
thrd_stop()
{
	wf_thrd *p;

	ITER(p, wf_thrd, &thrd_head) {
		(*p->give_up)(p->closure);
		p->flag &= ~(WF_THREAD_ERROR|WF_THREAD_RISKY);
			/* called only once */
	}
}

void
thrd_inspect(reqp)
wf_req *reqp;
{
	wf_thrd *p;

	ITER(p, wf_thrd, &thrd_head) {
		req_send(reqp, "thread %lx: vol= %s, dir= %s\n",
			 p, PROT(p->child_volp->name), PROT(p->dirp->name));
		req_send(reqp, "\tstep= %d, flag= %d, fh= {%lx, %lx, %lx, %lx}\n",
			 p->step, p->flag,
			 p->fhp->parent_vol, p->fhp->dir_id,
			 p->fhp->child_vol, p->fhp->file_id);
		req_send(reqp, "\tfname= %s, idle= %d, conn= %lx\n",
			 PROT(p->fname), p->idle, p->cp);
	}
}
