#!/usr/local/bin/perl
#
# Read $usg below.
# Author: Youki Kadobayashi <youki@wide.ad.jp>
# This file is part of WWFS.
#
require '/etc/csd.pl';

$usg = "sortvol.pl: Rearrange servers in volume description files.
Usage:	sortvol.pl [-d <gtr-directory>] <volume-files...>
	-d <gtr-directory>	alternative directory for GTR config files.
Files:	$WWFSDIR/etc/gtr.conf	for server preference
	$WWFSDIR/etc/gtr.txt	for GTR resource records
";
unshift(@INC, "$WWFSDIR/lib");
require 'gtr.pl';
require 'getopts.pl';

do Getopts('d:');
$opt_d = "$WWFSDIR/etc" unless $opt_d;
$gtr'conf = "$opt_d/gtr.conf";
$gtr'txt = "$opt_d/gtr.txt";

if ($#ARGV < $[) {
    die $usg;
}

foreach $file (@ARGV) {
    unless (open(IN, "$file")) {
	print "$file: $!\n";
	next;
    }
    rename($file, $file . '.bak');
    open(OUT, ">$file");
    select(OUT);

    while (<IN>) {
	chop;
	/ftp-server:\s+/ && do {
	    $host = $';
	};
	/volume-id:/ && undef $host;
	/ufs-directory:/ && undef $host;
	/description:/ && undef $host;
	unless (defined $host) {
	    print $_, "\n";
	    next;
	}
	$lines{$host} .= $_ . "\n";
    }
    close(IN);

    @sorted = &gtr'sort(keys %lines);
    foreach $host (@sorted) {
	print $lines{$host};
    }
    close(OUT);
    undef %lines;
}

# Local Variables:
# mode: perl
# End:
