#!/usr/local/bin/perl
#
#	renamesym.pl - rename symbol name in a object file.
#		5minutes hack for splicing libc.
# Author: Youki Kadobayashi <youki@ics.es.osaka-u.ac.jp>
#
# Usage: renamesym.pl [-l libraryfile] [-c] [-d] [-p] [-v] [-S sections]
#	libraryfile can be either archive (.a) or object (.o).
#	-c for capitalizing initial letter only.
#	-d for debug! Children under age of 18 should not use this option.
#	-p to preserve original object files with extension ".orig".
#	-v for verbose.
#	-S for other section letters.

require 'getopts.pl';
do Getopts('l:cdpvS:');

&usage if ((@ARGV) == 0);

sub usage {
    system("head -14 $0");
}

$lib = $opt_l ? $opt_l : "/usr/lib/libc.a";
$from_archive = 1 if ($lib =~ /\.a$/);
$is_mips = 1 if (-e "/usr/lib/cmplrs");  # I guess you have MIPS cc
$sections = $opt_S ? "[$opt_S]" : "^[0-9a-fA-F]+ [ADGIT]";
print "NM target sections are: $sections\n" if $opt_v;

foreach $func (@ARGV) {
    @objs = &search_library($lib, $func) if $from_archive;
    next if !defined(@objs);

    foreach $obj (@objs) {
	&rename_symbol($obj, $func);
    }
}

sub rename_symbol {
    local ($obj, $func) = @_;

    $func = "_" . $func unless $is_mips;
    $Func = $func;
    if ($opt_c) {
	$Func =~ /([^a-z]?)([a-z])(.*)$/;
	$foo = $2;
	$foo =~ tr/a-z/A-Z/;
	$Func = $1 . $foo . $3;
    } else {
	$Func =~ tr/a-z/A-Z/;
    }

    if ($from_archive) {
	if (-f "$obj.o" && $modified_files =~ /:$obj.o:/) {
	    # do not extract this file again!
	    print STDERR "Using previously modified $obj.o.\n" if $opt_v;
	} else {
	    print STDERR "Extracting from $lib...\n" if $opt_v;
	    # ar x /usr/lib/libc.a foo.o
	    $line = "ar x $lib $obj.o";
	    system($line);
	    $modified_files .= ":$obj.o:";
	}
    } else {
	$obj = $lib;
    }
    
    print STDERR "Rename $func() to $Func().\n" if $opt_v;
    open(OBJ, "$obj.o") || die;
    open(OUT, ">new.$obj.o") || die;
    while (<OBJ>) {
	# sed -e 's/foo/Foo/g' < foo.o > new.foo.o
	s/(\W)$func\000/\1$Func\000/g;
	print OUT;
    }
    close(OBJ);
    close(OUT);
    rename("$obj.o", "$obj.o.orig");
    rename("new.$obj.o", "$obj.o");
     
    if ($opt_d) {
	system <<"EOF";
	echo "================ $obj.o"
	od -c $obj.o > /tmp/b
	od -c $obj.o.orig > /tmp/a
	diff -c /tmp/[ab]
	rm -f /tmp/[ab]
EOF
    ;
    }
    unlink("$obj.o.orig") unless $opt_p;
}

sub search_library {
    local ($lib, $func) = @_;
    local ($obj, $matches);
    local (@found);

    print STDERR "Searching $lib for $func()...\n" if $opt_v;
    open(NM, "nm $lib |");
    while (<NM>) {
	/^(\S+)\.o:/ && ($obj = $1);
	if (/$func$/) {
	    $matches .= $_;
	    if (/$sections $func/) {
		push(found, $obj);
		last unless $opt_S;
	    }
	}
    }
    close(NM);			# yes, we love broken pipe!
    print STDERR "Found $func() in @found.\n" if $opt_v && @found;
    if (! scalar(@found)) {
	print STDERR "Could not find $func() in $lib.\n";
	print STDERR "Here's similar symbols in $lib:\n", $matches;
    }
    return @found;
}

# exit

# Local Variables:
# mode: perl
# End:
