#!/usr/local/bin/perl
#
# Read $usg below.
# Author: Youki Kadobayashi <youki@wide.ad.jp>
# This file is part of WWFS.
#
require '/etc/csd.pl';

$usg = "vol2html.pl: Convert volume files into HTML pages.
Usage:	vol2html [-l <librarian>] [-t <table>] <volume-files...>
	-l	your favorite librarian (default: youki)
	-t	translation table file (default: $WWFSDIR/etc/jtod.txt)
";

require 'getopts.pl';
do Getopts('l:t:');
$lib = $opt_l ? $opt_l : 'youki';
$tab = $opt_t ? $opt_t : "$WWFSDIR/etc/jtod.txt";

if ($#ARGV < $[) {
    die $usg;
}

# jargon to description mapping file.
open(TAB, "$tab") || die;
while (<TAB>) {
	chop;
	($key, $value) = split(/\s+/, $_, 2);
	$jtod{$key} = $value;
}
close(TAB);

foreach $file (@ARGV) {
	$vol = $file;
	$vol =~ s/.vol$//;
	open(IN, $file) || die;
	while (<IN>) {
		chop;
		if (/^description:\s*/i) {
			$description = $';
		};
		if (/^#category-$lib:\s*/i) {
			@c = split(/\s*,\s*/, $');
		}
	}
	foreach $key (@c) {
		$text{$key} .= "<LI> <A HREF=\"wwfs:/$vol\">$vol</A>, $description\n";
	}
	close(IN);
}

foreach $category (keys %text) {
	@cat = split(/\//, $category);
	if (! defined $page{$cat[0]}) {
		# try to convert jargon into usual language
		$title = $jtod{$cat[0]};
		if (! defined $title) {
			print "Using jargon word \"$cat[0]\" as is\n";
			$title = $cat[0];
		}
		$page{$cat[0]} = "<TITLE>$title</TITLE>\n<H1>$title</H1>\n\n";
	}
	if (! $done{$category}) {
		$done{$category} = 1;
		# try to convert jargon into usual language
		$title = $jtod{$cat[1]};
		if (! defined $title) {
			print "Using jargon word \"$cat[1]\" as is\n";
			$title = $cat[1];
		}
		$page{$cat[0]} .= "<H2>$title</H2>\n<UL>\n";
		$page{$cat[0]} .= $text{$category};
		$page{$cat[0]} .= "</UL>\n\n";
	}
}

foreach $name (keys %page) {
	open(FILE, "> vol_${name}.html") || die;
	print FILE $page{$name};
	print FILE "<!-- vol2html -l $lib ";
	print FILE "at " . `date` . "-->\n";
	close(FILE);
}

#	librarian at your choice.
#	category-youki: free/emacs, lang/lisp
#	category-suguru: free/gnu, lang/elisp
#
#	jargon to description mapping table "jtod.txt"
#		free --> Free Software
#		lang --> Language
#		emacs --> Emacs
#
#	vol_"eval &jtod($category, 1)".html
#
#	<TITLE>&jtod($category, 1)</TITLE>
#	<H1>&jtod($category, 1)</H1>
#
#	<H2>&jtod($category, 2)</H2>
#	<UL>
#	<LI> <A HREF="wwfs:/$volume">$volume</A>, $description
#	</UL>

# Local Variables:
# mode: cperl
# cperl-indent-level: 8
# cperl-continued-statement-offset: 8
# End:
